@@ -1,6 +1,6 @@
 /*
    neon SSL/TLS support using GNU TLS
-   Copyright (C) 2002-2005, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2002-2008, Joe Orton <joe@manyfish.co.uk>
    Copyright (C) 2004, Aleix Conchillo Flaque <aleix@member.fsf.org>
 
    This library is free software; you can redistribute it and/or
@@ -17,13 +17,10 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA
-
 */
 
 #include "config.h"
 
-#include <sys/mman.h>
-#include <sys/stat.h>
 #include <sys/types.h>
 
 #ifdef HAVE_STRING_H
@@ -33,14 +30,28 @@
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
+#include <errno.h>
 
 #include <gnutls/gnutls.h>
 #include <gnutls/pkcs12.h>
 
+#ifdef NE_HAVE_TS_SSL
+#include <errno.h>
+#include <pthread.h>
+#include <gcrypt.h>
+GCRY_THREAD_OPTION_PTHREAD_IMPL;
+#else
+#include <gcrypt.h>
+#endif
+
+#ifdef HAVE_ICONV
+#include <iconv.h>
+#endif
+
 #include "ne_ssl.h"
 #include "ne_string.h"
 #include "ne_session.h"
-#include "ne_i18n.h"
+#include "ne_internal.h"
 
 #include "ne_private.h"
 #include "ne_privssl.h"
@@ -60,6 +71,7 @@ struct ne_ssl_certificate_s {
 struct ne_ssl_client_cert_s {
     gnutls_pkcs12 p12;
     int decrypted; /* non-zero if successfully decrypted. */
+    int keyless;
     ne_ssl_certificate cert;
     gnutls_x509_privkey pkey;
     char *friendly_name;
@@ -86,6 +98,132 @@ static int oid_find_highest_index(gnutls_x509_crt cert, int subject, const char
     return idx - 1;
 }
 
+#ifdef HAVE_GNUTLS_X509_DN_GET_RDN_AVA
+/* New-style RDN handling introduced in GnuTLS 1.7.x. */
+
+#ifdef HAVE_ICONV
+static void convert_dirstring(ne_buffer *buf, const char *charset, 
+                              gnutls_datum *data)
+{
+    iconv_t id = iconv_open("UTF-8", charset);
+    size_t inlen = data->size, outlen = buf->length - buf->used;
+    char *inbuf = (char *)data->data;
+    char *outbuf = buf->data + buf->used - 1;
+    
+    if (id == (iconv_t)-1) {
+        char err[128], err2[128];
+
+        ne_snprintf(err, sizeof err, "[unprintable in %s: %s]",
+                    charset, ne_strerror(errno, err2, sizeof err2));
+        ne_buffer_zappend(buf, err);
+        return;
+    }
+    
+    ne_buffer_grow(buf, buf->used + 64);
+    
+    while (inlen && outlen 
+           && iconv(id, &inbuf, &inlen, &outbuf, &outlen) == 0)
+        ;
+    
+    iconv_close(id);
+    buf->used += buf->length - buf->used - outlen;
+    buf->data[buf->used - 1] = '\0';
+}
+#endif
+
+/* From section 11.13 of the Dubuisson ASN.1 bible: */
+#define TAG_UTF8 (12)
+#define TAG_PRINTABLE (19)
+#define TAG_T61 (20)
+#define TAG_IA5 (22)
+#define TAG_VISIBLE (26)
+#define TAG_UNIVERSAL (28)
+#define TAG_BMP (30)
+
+static void append_dirstring(ne_buffer *buf, gnutls_datum *data, unsigned long tag)
+{
+    switch (tag) {
+    case TAG_UTF8:
+    case TAG_IA5:
+    case TAG_PRINTABLE:
+    case TAG_VISIBLE:
+        ne_buffer_append(buf, (char *)data->data, data->size);
+        break;
+#ifdef HAVE_ICONV
+    case TAG_T61:
+        convert_dirstring(buf, "ISO-8859-1", data);
+        break;
+    case TAG_BMP:
+        convert_dirstring(buf, "UCS-2BE", data);
+        break;
+#endif
+    default: {
+        char tmp[128];
+        ne_snprintf(tmp, sizeof tmp, _("[unprintable:#%lu]"), tag);
+        ne_buffer_zappend(buf, tmp);
+    } break;
+    }
+}
+
+/* OIDs to not include in readable DNs by default: */
+#define OID_emailAddress "1.2.840.113549.1.9.1"
+#define OID_commonName "2.5.4.3"
+
+#define CMPOID(a,o) ((a)->oid.size == sizeof(o)                        \
+                     && memcmp((a)->oid.data, o, strlen(o)) == 0)
+
+char *ne_ssl_readable_dname(const ne_ssl_dname *name)
+{
+    gnutls_x509_dn_t dn;
+    int ret, rdn = 0, flag = 0;
+    ne_buffer *buf;
+    gnutls_x509_ava_st val;
+
+    if (name->subject)
+        ret = gnutls_x509_crt_get_subject(name->cert, &dn);
+    else
+        ret = gnutls_x509_crt_get_issuer(name->cert, &dn);
+    
+    if (ret)
+        return ne_strdup(_("[unprintable]"));
+
+    buf = ne_buffer_create();
+    
+    /* Find the highest rdn... */
+    while (gnutls_x509_dn_get_rdn_ava(dn, rdn++, 0, &val) == 0)
+        ;        
+
+    /* ..then iterate back to the first: */
+    while (--rdn >= 0) {
+        int ava = 0;
+
+        /* Iterate through all AVAs for multivalued AVAs; better than
+         * ne_openssl can do! */
+        do {
+            ret = gnutls_x509_dn_get_rdn_ava(dn, rdn, ava, &val);
+
+            /* If the *only* attribute to append is the common name or
+             * email address, use it; otherwise skip those
+             * attributes. */
+            if (ret == 0 && val.value.size > 0
+                && ((!CMPOID(&val, OID_emailAddress)
+                     && !CMPOID(&val, OID_commonName))
+                    || (buf->used == 1 && rdn == 0))) {
+                flag = 1;
+                if (buf->used > 1) ne_buffer_append(buf, ", ", 2);
+
+                append_dirstring(buf, &val.value, val.value_tag);
+            }
+            
+            ava++;
+        } while (ret == 0);
+    }
+
+    return ne_buffer_finish(buf);
+}
+
+#else /* !HAVE_GNUTLS_X509_DN_GET_RDN_AVA */
+
 /* Appends the value of RDN with given oid from certitifcate x5
  * subject (if subject is non-zero), or issuer DN to buffer 'buf': */
 static void append_rdn(ne_buffer *buf, gnutls_x509_crt x5, int subject, const char *oid)
@@ -114,13 +252,9 @@ static void append_rdn(ne_buffer *buf, gnutls_x509_crt x5, int subject, const ch
     }
 }
 
-
 char *ne_ssl_readable_dname(const ne_ssl_dname *name)
 {
     ne_buffer *buf = ne_buffer_create();
-#if 0
-    /* this code can be used once there is a released version of GnuTLS
-     * with fixed _get_dn_oid functions */
     int ret, idx = 0;
 
     do {
@@ -136,29 +270,35 @@ char *ne_ssl_readable_dname(const ne_ssl_dname *name)
             idx++;
         }
     } while (ret != GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE);
-#else
-
-#define APPEND_RDN(x) append_rdn(buf, name->cert, name->subject, GNUTLS_OID_##x)
-
-    APPEND_RDN(X520_ORGANIZATIONAL_UNIT_NAME);
-    APPEND_RDN(X520_ORGANIZATION_NAME);
-    APPEND_RDN(X520_LOCALITY_NAME);
-    APPEND_RDN(X520_STATE_OR_PROVINCE_NAME);
-    APPEND_RDN(X520_COUNTRY_NAME);
-
-    if (buf->used == 1) APPEND_RDN(X520_COMMON_NAME);
-    if (buf->used == 1) APPEND_RDN(PKCS9_EMAIL);
-
-#undef APPEND_RDN
-#endif
 
     return ne_buffer_finish(buf);
 }
+#endif /* HAVE_GNUTLS_X509_DN_GET_RDN_AVA */
 
 int ne_ssl_dname_cmp(const ne_ssl_dname *dn1, const ne_ssl_dname *dn2)
 {
-#warning incomplete
-    return 1;
+    char c1[1024], c2[1024];
+    size_t s1 = sizeof c1, s2 = sizeof c2;
+    int ret;
+
+    if (dn1->subject)
+        ret = gnutls_x509_crt_get_dn(dn1->cert, c1, &s1);
+    else
+        ret = gnutls_x509_crt_get_issuer_dn(dn1->cert, c1, &s1);
+    if (ret)
+        return 1;
+
+    if (dn2->subject)
+        ret = gnutls_x509_crt_get_dn(dn2->cert, c2, &s2);
+    else
+        ret = gnutls_x509_crt_get_issuer_dn(dn2->cert, c2, &s2);
+    if (ret)
+        return -1;
+    
+    if (s1 != s2)
+        return s2 - s1;
+
+    return memcmp(c1, c2, s1);
 }
 
 void ne_ssl_clicert_free(ne_ssl_client_cert *cc)
@@ -174,48 +314,23 @@ void ne_ssl_clicert_free(ne_ssl_client_cert *cc)
     ne_free(cc);
 }
 
-void ne_ssl_cert_validity(const ne_ssl_certificate *cert,
-                          char *from, char *until)
+void ne_ssl_cert_validity_time(const ne_ssl_certificate *cert,
+                               time_t *from, time_t *until)
 {
-#warning FIXME strftime not portable
     if (from) {
-        time_t t = gnutls_x509_crt_get_activation_time(cert->subject);
-        strftime(from, NE_SSL_VDATELEN, "%b %d %H:%M:%S %Y %Z", localtime(&t));
+        *from = gnutls_x509_crt_get_activation_time(cert->subject);
     }
     if (until) {
-        time_t t = gnutls_x509_crt_get_expiration_time(cert->subject);
-        strftime(until, NE_SSL_VDATELEN, "%b %d %H:%M:%S %Y %Z", localtime(&t));
+        *until = gnutls_x509_crt_get_expiration_time(cert->subject);
     }
 }
 
-/* Return non-zero if hostname from certificate (cn) matches hostname
- * used for session (hostname).  (Wildcard matching is no longer
- * mandated by RFC3280, but certs are deployed which use wildcards) */
-static int match_hostname(char *cn, const char *hostname)
-{
-    const char *dot;
-    NE_DEBUG(NE_DBG_SSL, "Match %s on %s...\n", cn, hostname);
-    dot = strchr(hostname, '.');
-    if (dot == NULL) {
-	char *pnt = strchr(cn, '.');
-	/* hostname is not fully-qualified; unqualify the cn. */
-	if (pnt != NULL) {
-	    *pnt = '\0';
-	}
-    }
-    else if (strncmp(cn, "*.", 2) == 0) {
-	hostname = dot + 1;
-	cn += 2;
-    }
-    return !strcasecmp(cn, hostname);
-}
-
 /* Check certificate identity.  Returns zero if identity matches; 1 if
  * identity does not match, or <0 if the certificate had no identity.
  * If 'identity' is non-NULL, store the malloc-allocated identity in
  * *identity.  If 'server' is non-NULL, it must be the network address
  * of the server in use, and identity must be NULL. */
-static int check_identity(const char *hostname, gnutls_x509_crt cert,
+static int check_identity(const ne_uri *server, gnutls_x509_crt cert,
                           char **identity)
 {
     char name[255];
@@ -223,17 +338,69 @@ static int check_identity(const char *hostname, gnutls_x509_crt cert,
     int ret, seq = 0;
     int match = 0, found = 0;
     size_t len;
+    const char *hostname;
+    
+    hostname = server ? server->host : "";
 
     do {
-        len = sizeof name;
+        len = sizeof name - 1;
         ret = gnutls_x509_crt_get_subject_alt_name(cert, seq, name, &len,
                                                    &critical);
         switch (ret) {
         case GNUTLS_SAN_DNSNAME:
+            name[len] = '\0';
             if (identity && !found) *identity = ne_strdup(name);
-            match = match_hostname(name, hostname);
+            match = ne__ssl_match_hostname(name, hostname);
             found = 1;
             break;
+        case GNUTLS_SAN_IPADDRESS: {
+            ne_inet_addr *ia;
+            if (len == 4)
+                ia = ne_iaddr_make(ne_iaddr_ipv4, (unsigned char *)name);
+            else if (len == 16)
+                ia = ne_iaddr_make(ne_iaddr_ipv6, (unsigned char *)name);
+            else 
+                ia = NULL;
+            if (ia) {
+                char buf[128];
+                
+                match = strcmp(hostname, 
+                               ne_iaddr_print(ia, buf, sizeof buf)) == 0;
+                if (identity) *identity = ne_strdup(buf);
+                found = 1;
+                ne_iaddr_free(ia);
+            } else {
+                NE_DEBUG(NE_DBG_SSL, "iPAddress name with unsupported "
+                         "address type (length %" NE_FMT_SIZE_T "), skipped.\n",
+                         len);
+            }
+        } break;
+        case GNUTLS_SAN_URI: {
+            ne_uri uri;
+            
+            name[len] = '\0';
+            
+            if (ne_uri_parse(name, &uri) == 0 && uri.host && uri.scheme) {
+                ne_uri tmp;
+                
+                if (identity && !found) *identity = ne_strdup(name);
+                found = 1;
+                
+                if (server) {
+                    /* For comparison purposes, all that matters is
+                     * host, scheme and port; ignore the rest. */
+                    memset(&tmp, 0, sizeof tmp);
+                    tmp.host = uri.host;
+                    tmp.scheme = uri.scheme;
+                    tmp.port = uri.port;
+                    
+                    match = ne_uri_cmp(server, &tmp) == 0;
+                }
+            }
+            
+            ne_uri_free(&uri);
+        } break;
+
         default:
             break;
         }
@@ -252,14 +419,17 @@ static int check_identity(const char *hostname, gnutls_x509_crt cert,
                                                 seq, 0, name, &len);
             if (ret == 0) {
                 if (identity) *identity = ne_strdup(name);
-                match = match_hostname(name, hostname);
+                match = ne__ssl_match_hostname(name, hostname);
             }
         } else {
             return -1;
         }
     }
 
-    NE_DEBUG(NE_DBG_SSL, "Identity match: %s\n", match ? "good" : "bad");
+    if (*hostname)
+        NE_DEBUG(NE_DBG_SSL, "ssl: Identity match for '%s': %s\n", hostname, 
+                 match ? "good" : "bad");
+
     return match ? 0 : 1;
 }
 
@@ -274,7 +444,7 @@ static ne_ssl_certificate *populate_cert(ne_ssl_certificate *cert,
     cert->issuer = NULL;
     cert->subject = x5;
     cert->identity = NULL;
-    check_identity("", x5, &cert->identity);
+    check_identity(NULL, x5, &cert->identity);
     return cert;
 }
 
@@ -286,7 +456,7 @@ static gnutls_x509_crt x509_crt_copy(gnutls_x509_crt src)
     gnutls_datum tmp;
     gnutls_x509_crt dest;
     
-    if (gnutls_x509_crt_init(&dest) == 0) {
+    if (gnutls_x509_crt_init(&dest) != 0) {
         return NULL;
     }
 
@@ -319,13 +489,18 @@ static ne_ssl_client_cert *dup_client_cert(const ne_ssl_client_cert *cc)
     ne_ssl_client_cert *newcc = ne_calloc(sizeof *newcc);
 
     newcc->decrypted = 1;
-
-    ret = gnutls_x509_privkey_init(&newcc->pkey);
-    if (ret != 0) goto dup_error;
-
-    ret = gnutls_x509_privkey_cpy(newcc->pkey, cc->pkey);
-    if (ret != 0) goto dup_error;
     
+    if (cc->keyless) {
+        newcc->keyless = 1;
+    }
+    else {
+        ret = gnutls_x509_privkey_init(&newcc->pkey);
+        if (ret != 0) goto dup_error;
+        
+        ret = gnutls_x509_privkey_cpy(newcc->pkey, cc->pkey);
+        if (ret != 0) goto dup_error;
+    }    
+
     newcc->cert.subject = x509_crt_copy(cc->cert.subject);
     if (!newcc->cert.subject) goto dup_error;
 
@@ -341,6 +516,51 @@ dup_error:
     return NULL;
 }    
 
+/* Callback invoked when the SSL server requests a client certificate.  */
+static int provide_client_cert(gnutls_session session,
+                               const gnutls_datum *req_ca_rdn, int nreqs,
+                               const gnutls_pk_algorithm *sign_algos,
+                               int sign_algos_length, gnutls_retr_st *st)
+{
+    ne_session *sess = gnutls_session_get_ptr(session);
+    
+    if (!sess) {
+        return GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER;
+    }
+
+    if (!sess->client_cert && sess->ssl_provide_fn) {
+        /* The dname array cannot be converted without better dname
+         * support from GNUTLS. */
+        sess->ssl_provide_fn(sess->ssl_provide_ud, sess,
+                             NULL, 0);
+    }
+
+    NE_DEBUG(NE_DBG_SSL, "In client cert provider callback.\n");
+
+    if (sess->client_cert) {
+        gnutls_certificate_type type = gnutls_certificate_type_get(session);
+        if (type == GNUTLS_CRT_X509) {
+            NE_DEBUG(NE_DBG_SSL, "Supplying client certificate.\n");
+
+            st->type = type;
+            st->ncerts = 1;
+            st->cert.x509 = &sess->client_cert->cert.subject;
+            st->key.x509 = sess->client_cert->pkey;
+            
+            /* tell GNU TLS not to deallocate the certs. */
+            st->deinit_all = 0;
+        } else {
+            return GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE;
+        }
+    } else {
+        NE_DEBUG(NE_DBG_SSL, "No client certificate supplied.\n");
+        sess->ssl_cc_requested = 1;
+        return GNUTLS_E_NO_CERTIFICATE_FOUND;
+    }
+
+    return 0;
+}
+
 void ne_ssl_set_clicert(ne_session *sess, const ne_ssl_client_cert *cc)
 {
     sess->client_cert = dup_client_cert(cc);
@@ -348,8 +568,12 @@ void ne_ssl_set_clicert(ne_session *sess, const ne_ssl_client_cert *cc)
 
 ne_ssl_context *ne_ssl_context_create(int flags)
 {
-    ne_ssl_context *ctx = ne_malloc(sizeof *ctx);
+    ne_ssl_context *ctx = ne_calloc(sizeof *ctx);
     gnutls_certificate_allocate_credentials(&ctx->cred);
+    if (flags == NE_SSL_CTX_CLIENT) {
+        gnutls_certificate_client_set_retrieve_function(ctx->cred,
+                                                        provide_client_cert);
+    }
     return ctx;
 }
 
@@ -364,18 +588,31 @@ int ne_ssl_context_keypair(ne_ssl_context *ctx,
 int ne_ssl_context_set_verify(ne_ssl_context *ctx, int required,
                               const char *ca_names, const char *verify_cas)
 {
+    ctx->verify = required;
     if (verify_cas) {
         gnutls_certificate_set_x509_trust_file(ctx->cred, verify_cas,
                                                GNUTLS_X509_FMT_PEM);
     }
-#warning argh
+    /* gnutls_certificate_send_x509_rdn_sequence in gnutls >= 1.2 can
+     * be used to *suppress* sending the CA names, but not control it,
+     * it seems. */
     return 0;
 }
 
+void ne_ssl_context_set_flag(ne_ssl_context *ctx, int flag, int value)
+{
+    /* SSLv2 not supported. */
+}
 
 void ne_ssl_context_destroy(ne_ssl_context *ctx)
 {
     gnutls_certificate_free_credentials(ctx->cred);
+    if (ctx->cache.client.data) {
+        ne_free(ctx->cache.client.data);
+    } else if (ctx->cache.server.key.data) {
+        gnutls_free(ctx->cache.server.key.data);
+        gnutls_free(ctx->cache.server.data.data);
+    }    
     ne_free(ctx);
 }
 
@@ -392,16 +629,16 @@ static ne_ssl_certificate *make_peers_chain(gnutls_session sock)
     }
     
     for (n = 0; n < count; n++) {
-        ne_ssl_certificate *cert = ne_malloc(sizeof *cert);
+        ne_ssl_certificate *cert;
         gnutls_x509_crt x5;
 
         if (gnutls_x509_crt_init(&x5) ||
             gnutls_x509_crt_import(x5, &certs[n], GNUTLS_X509_FMT_DER)) {
-            /* leak! */
+            ne_ssl_cert_free(top);
             return NULL;
         }
 
-        populate_cert(cert, x5);
+        cert = populate_cert(ne_malloc(sizeof *cert), x5);
         
         if (top == NULL) {
             current = top = cert;
@@ -420,6 +657,7 @@ static int check_certificate(ne_session *sess, gnutls_session sock,
 {
     time_t before, after, now = time(NULL);
     int ret, failures = 0;
+    ne_uri server;
 
     before = gnutls_x509_crt_get_activation_time(chain->subject);
     after = gnutls_x509_crt_get_expiration_time(chain->subject);
@@ -429,7 +667,11 @@ static int check_certificate(ne_session *sess, gnutls_session sock,
     else if (now > after)
         failures |= NE_SSL_EXPIRED;
 
-    ret = check_identity(sess->server.hostname, chain->subject, NULL);
+    memset(&server, 0, sizeof server);
+    ne_fill_server_uri(sess, &server);
+    ret = check_identity(&server, chain->subject, NULL);
+    ne_uri_free(&server);
+
     if (ret < 0) {
         ne_set_error(sess, _("Server certificate was missing commonName "
                              "attribute in subject name"));
@@ -447,7 +689,7 @@ static int check_certificate(ne_session *sess, gnutls_session sock,
     if (failures == 0) {
         ret = NE_OK;
     } else {
-#warning TODO: set up error string
+        ne__ssl_set_verify_err(sess, failures);
         ret = NE_ERROR;
         if (sess->ssl_verify_fn
             && sess->ssl_verify_fn(sess->ssl_verify_ud, failures, chain) == 0)
@@ -458,19 +700,29 @@ static int check_certificate(ne_session *sess, gnutls_session sock,
 }
 
 /* Negotiate an SSL connection. */
-int ne__negotiate_ssl(ne_request *req)
+int ne__negotiate_ssl(ne_session *sess)
 {
-    ne_session *const sess = ne_get_session(req);
     ne_ssl_context *const ctx = sess->ssl_context;
     ne_ssl_certificate *chain;
     gnutls_session sock;
 
     NE_DEBUG(NE_DBG_SSL, "Negotiating SSL connection.\n");
 
+    /* Pass through the hostname if SNI is enabled. */
+    ctx->hostname = 
+        sess->flags[NE_SESSFLAG_TLS_SNI] ? sess->server.hostname : NULL;
+
     if (ne_sock_connect_ssl(sess->socket, ctx, sess)) {
-	ne_set_error(sess, _("SSL negotiation failed: %s"),
-		     ne_sock_error(sess->socket));
-	return NE_ERROR;
+        if (sess->ssl_cc_requested) {
+            ne_set_error(sess, _("SSL negotiation failed, "
+                                 "client certificate was requested: %s"),
+                         ne_sock_error(sess->socket));
+        }
+        else {
+            ne_set_error(sess, _("SSL negotiation failed: %s"),
+                         ne_sock_error(sess->socket));
+        }
+        return NE_ERROR;
     }
 
     sock = ne__sock_sslsock(sess->socket);
@@ -481,11 +733,21 @@ int ne__negotiate_ssl(ne_request *req)
         return NE_ERROR;
     }
 
+    if (sess->server_cert && ne_ssl_cert_cmp(sess->server_cert, chain) == 0) {
+        /* Same cert as last time; presume OK.  This is not optimal as
+         * make_peers_chain() has already gone through and done the
+         * expensive DER parsing stuff for the whole chain by now. */
+        ne_ssl_cert_free(chain);
+        return NE_OK;
+    }
+
     if (check_certificate(sess, sock, chain)) {
         ne_ssl_cert_free(chain);
         return NE_ERROR;
     }
 
+    sess->server_cert = chain;
+
     return NE_OK;
 }
 
@@ -517,7 +779,11 @@ void ne_ssl_context_trustcert(ne_ssl_context *ctx, const ne_ssl_certificate *cer
 
 void ne_ssl_trust_default_ca(ne_session *sess)
 {
-#warning incomplete
+#ifdef NE_SSL_CA_BUNDLE
+    gnutls_certificate_set_x509_trust_file(sess->ssl_context->cred,
+                                           NE_SSL_CA_BUNDLE,
+                                           GNUTLS_X509_FMT_PEM);
+#endif
 }
 
 /* Read the contents of file FILENAME into *DATUM. */
@@ -538,12 +804,15 @@ static int read_to_datum(const char *filename, gnutls_datum *datum)
     }
 
     if (!feof(f)) {
+        fclose(f);
         ne_buffer_destroy(buf);
         return -1;
     }
     
+    fclose(f);
+
     datum->size = ne_buffer_size(buf);
-    datum->data = ne_buffer_finish(buf);
+    datum->data = (unsigned char *)ne_buffer_finish(buf);
     return 0;
 }
 
@@ -566,14 +835,14 @@ static int pkcs12_parse(gnutls_pkcs12 p12, gnutls_x509_privkey *pkey,
         ret = gnutls_pkcs12_get_bag(p12, i, bag);
         if (ret < 0) continue;
 
-        gnutls_pkcs12_bag_decrypt(bag, password == NULL ? "" : password);
+        gnutls_pkcs12_bag_decrypt(bag, password);
 
         for (j = 0; ret == 0 && j < gnutls_pkcs12_bag_get_count(bag); ++j) {
             gnutls_pkcs12_bag_type type;
             gnutls_datum data;
 
             if (friendly_name && *friendly_name == NULL) {
-                char *name;
+                char *name = NULL;
                 gnutls_pkcs12_bag_get_friendly_name(bag, j, &name);
                 if (name) {
                     if (name[0] == '.') name++; /* weird GnuTLS bug? */
@@ -650,9 +919,13 @@ ne_ssl_client_cert *ne_ssl_clicert_read(const char *filename)
         return NULL;
     }
 
-    ret = pkcs12_parse(p12, &pkey, &cert, &friendly_name, NULL);
-
     if (gnutls_pkcs12_verify_mac(p12, "") == 0) {
+        if (pkcs12_parse(p12, &pkey, &cert, &friendly_name, "") != 0
+            || !cert || !pkey) {
+            gnutls_pkcs12_deinit(p12);
+            return NULL;
+        }
+
         cc = ne_calloc(sizeof *cc);
         cc->pkey = pkey;
         cc->decrypted = 1;
@@ -662,18 +935,38 @@ ne_ssl_client_cert *ne_ssl_clicert_read(const char *filename)
         cc->p12 = NULL;
         return cc;
     } else {
-        if (ret == 0) {
-            cc = ne_calloc(sizeof *cc);
-            cc->friendly_name = friendly_name;
-            cc->p12 = p12;
-            return cc;
-        } else {
-            gnutls_pkcs12_deinit(p12);
-            return NULL;
-        }
+        /* TODO: calling pkcs12_parse() here to find the friendly_name
+         * seems to break horribly.  */
+        cc = ne_calloc(sizeof *cc);
+        cc->p12 = p12;
+        return cc;
     }
 }
 
+ne_ssl_client_cert *ne__ssl_clicert_exkey_import(const unsigned char *der,
+                                                 size_t der_len)
+{
+    ne_ssl_client_cert *cc;
+    gnutls_x509_crt x5;
+    gnutls_datum datum;
+
+    datum.data = (unsigned char *)der;
+    datum.size = der_len;    
+
+    if (gnutls_x509_crt_init(&x5) 
+        || gnutls_x509_crt_import(x5, &datum, GNUTLS_X509_FMT_DER)) {
+        NE_DEBUG(NE_DBG_SSL, "ssl: crt_import failed.\n");
+        return NULL;
+    }
+    
+    cc = ne_calloc(sizeof *cc);
+    cc->keyless = 1;
+    cc->decrypted = 1;
+    populate_cert(&cc->cert, x5);
+
+    return cc;    
+}
+
 int ne_ssl_clicert_encrypted(const ne_ssl_client_cert *cc)
 {
     return !cc->decrypted;
@@ -692,6 +985,12 @@ int ne_ssl_clicert_decrypt(ne_ssl_client_cert *cc, const char *password)
     ret = pkcs12_parse(cc->p12, &pkey, &cert, NULL, password);
     if (ret < 0)
         return ret;
+    
+    if (!cert || (!pkey && !cc->keyless)) {
+        if (cert) gnutls_x509_crt_deinit(cert);
+        if (pkey) gnutls_x509_privkey_deinit(pkey);
+        return -1;
+    }
 
     gnutls_pkcs12_deinit(cc->p12);
     populate_cert(&cc->cert, cert);
@@ -736,7 +1035,7 @@ ne_ssl_certificate *ne_ssl_cert_read(const char *filename)
 int ne_ssl_cert_write(const ne_ssl_certificate *cert, const char *filename)
 {
     unsigned char buffer[10*1024];
-    int len = sizeof buffer;
+    size_t len = sizeof buffer;
 
     FILE *fp = fopen(filename, "w");
 
@@ -835,8 +1134,9 @@ char *ne_ssl_cert_export(const ne_ssl_certificate *cert)
 
 int ne_ssl_cert_digest(const ne_ssl_certificate *cert, char *digest)
 {
-    int j, len = 20;
     char sha1[20], *p;
+    int j;
+    size_t len = sizeof sha1;
 
     if (gnutls_x509_crt_get_fingerprint(cert->subject, GNUTLS_DIG_SHA,
                                         sha1, &len) < 0)
@@ -851,3 +1151,25 @@ int ne_ssl_cert_digest(const ne_ssl_certificate *cert, char *digest)
     *--p = '\0';
     return 0;
 }
+
+int ne__ssl_init(void)
+{
+#ifdef NE_HAVE_TS_SSL
+    gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
+#endif
+    gcry_control(GCRYCTL_ENABLE_QUICK_RANDOM, 0);
+    return gnutls_global_init();
+}
+
+void ne__ssl_exit(void)
+{
+    /* No way to unregister the thread callbacks.  Doomed. */
+#if LIBGNUTLS_VERSION_MAJOR > 1 || LIBGNUTLS_VERSION_MINOR > 3 \
+    || (LIBGNUTLS_VERSION_MINOR == 3 && LIBGNUTLS_VERSION_PATCH >= 3)
+    /* It's safe to call gnutls_global_deinit() here only with
+     * gnutls >= 1.3., since older versions don't refcount and
+     * doing so would prevent any other use of gnutls within
+     * the process. */
+    gnutls_global_deinit();
+#endif
+}
