@@ -1,6 +1,6 @@
 /* 
    Authentication tests
-   Copyright (C) 2001-2007, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2001-2008, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -148,7 +148,10 @@ static int basic(void)
         /* multi-header case 3 */
         "WWW-Authenticate: FooBarChall foo=\"bar\"\r\n"
         CHAL_WALLY "\r\n"
-        "WWW-Authenticate: BarFooScheme bar=\"foo\""
+        "WWW-Authenticate: BarFooScheme bar=\"foo\"",
+
+        /* quoting test; fails to handle scheme properly with <= 0.28.2. */
+        "WWW-Authenticate: Basic realm=\"WallyWorld\" , BarFooScheme"
     };
     size_t n;
     
@@ -368,25 +371,29 @@ static void dup_header(char *header)
 }
 
 struct digest_parms {
-    const char *realm, *nonce, *opaque;
+    const char *realm, *nonce, *opaque, *domain;
     int rfc2617;
     int send_ainfo;
     int md5_sess;
     int proxy;
     int send_nextnonce;
     int num_requests;
+    int stale;
     enum digest_failure {
         fail_not,
         fail_bogus_alg,
+        fail_req0_stale,
         fail_omit_qop,
         fail_omit_realm,
         fail_omit_nonce,
         fail_ai_bad_nc,
+        fail_ai_bad_nc_syntax,
         fail_ai_bad_digest,
         fail_ai_bad_cnonce,
         fail_ai_omit_cnonce,
         fail_ai_omit_digest,
-        fail_ai_omit_nc
+        fail_ai_omit_nc,
+        fail_outside_domain
     } failure;
 };
 
@@ -461,10 +468,8 @@ static int check_digest(struct digest_state *state, struct digest_parms *parms)
 
     make_digest(state, parms, 0, digest);
 
-    NE_DEBUG(NE_DBG_HTTP, "Got digest: %s, expected: %s\n", 
-             state->digest, digest);
-
-    ONCMP(digest, state->digest, "Digest response", "request-digest");
+    ONV(strcmp(digest, state->digest),
+        ("bad digest; expected %s got %s", state->digest, digest));
 
     return OK;
 }
@@ -543,7 +548,8 @@ static int verify_digest_header(struct digest_state *state,
 
     DIGCMP(realm);
     DIGCMP(username);
-    DIGCMP(uri);
+    if (!parms->domain)
+        DIGCMP(uri);
     DIGCMP(nonce);
     DIGCMP(opaque);
     DIGCMP(algorithm);
@@ -555,6 +561,9 @@ static int verify_digest_header(struct digest_state *state,
     if (newstate.cnonce) {
         state->cnonce = ne_strdup(newstate.cnonce);
     }
+    if (parms->domain) {
+        state->uri = ne_strdup(newstate.uri);
+    }
 
     ONN("no digest param given", !state->digest);
 
@@ -577,7 +586,9 @@ static char *make_authinfo_header(struct digest_state *state,
         make_digest(state, parms, 1, digest);
     }
 
-    if (parms->failure == fail_ai_bad_nc) {
+    if (parms->failure == fail_ai_bad_nc_syntax) {
+        ncval = "zztop";
+    } else if (parms->failure == fail_ai_bad_nc) {
         ncval = "999";
     } else {
         ncval = state->ncval;
@@ -642,6 +653,15 @@ static char *make_digest_header(struct digest_state *state,
         ne_buffer_concat(buf, "opaque=\"", parms->opaque, "\", ", NULL);
     }
 
+    if (parms->domain) {
+        ne_buffer_concat(buf, "domain=\"", parms->domain, "\", ", NULL);
+    }
+
+    if (parms->failure == fail_req0_stale
+        || parms->stale == parms->num_requests) {
+        ne_buffer_concat(buf, "stale='true', ", NULL);
+    }
+
     ne_buffer_concat(buf, "nonce=\"", state->nonce, "\"", NULL);
 
     return ne_buffer_finish(buf);
@@ -654,7 +674,12 @@ static int serve_digest(ne_socket *sock, void *userdata)
     struct digest_state state;
     char resp[NE_BUFSIZ];
     
-    state.uri = parms->proxy ? "http://www.example.com/fish" : "/fish";
+    if (parms->proxy)
+        state.uri = "http://www.example.com/fish";
+    else if (parms->domain)
+        state.uri = "/fish/0";
+    else
+        state.uri = "/fish";
     state.method = "GET";
     state.realm = parms->realm;
     state.nonce = parms->nonce;
@@ -667,6 +692,11 @@ static int serve_digest(ne_socket *sock, void *userdata)
 
     state.cnonce = state.digest = state.ncval = NULL;
 
+    parms->num_requests += parms->stale ? 1 : 0;
+
+    NE_DEBUG(NE_DBG_HTTP, ">>>> Response sequence begins, %d requests.\n",
+             parms->num_requests);
+
     want_header = parms->proxy ? "Proxy-Authorization" : "Authorization";
     digest_hdr = NULL;
     got_header = dup_header;
@@ -685,14 +715,43 @@ static int serve_digest(ne_socket *sock, void *userdata)
 
     SEND_STRING(sock, resp);
 
+    /* Give up now if we've sent a challenge which should force the
+     * client to fail immediately: */
+    if (parms->failure == fail_bogus_alg
+        || parms->failure == fail_req0_stale) {
+        return OK;
+    }
+
     do {
+        digest_hdr = NULL;
         CALL(discard_request(sock));
 
-        ONN("no Authorization header sent", digest_hdr == NULL);
-        
-        CALL(verify_digest_header(&state, parms, digest_hdr));
-        
-        if (parms->send_ainfo) {
+        if (digest_hdr && parms->domain && (parms->num_requests & 1) != 0) {
+            SEND_STRING(sock, "HTTP/1.1 400 Used Auth Outside Domain\r\n\r\n");
+            return OK;
+        }
+        else if (digest_hdr == NULL && parms->domain && (parms->num_requests & 1) != 0) {
+            /* Do nothing. */
+            NE_DEBUG(NE_DBG_HTTP, "No Authorization header sent, good.\n");
+        }
+        else {
+            ONN("no Authorization header sent", digest_hdr == NULL);
+
+            CALL(verify_digest_header(&state, parms, digest_hdr));
+        }
+
+        if (parms->num_requests == parms->stale) {
+            state.nonce = ne_concat("stale-", state.nonce, NULL);
+            state.nc = 1;
+
+            ne_snprintf(resp, sizeof resp,
+                        "HTTP/1.1 %d Auth Denied\r\n"
+                        "%s\r\n"
+                        "Content-Length: 0\r\n" "\r\n",
+                        parms->proxy ? 407 : 401,
+                        make_digest_header(&state, parms));
+        }
+        else if (parms->send_ainfo) {
             char *ai = make_authinfo_header(&state, parms);
             
             ne_snprintf(resp, sizeof resp,
@@ -711,7 +770,7 @@ static int serve_digest(ne_socket *sock, void *userdata)
         SEND_STRING(sock, resp);
 
         NE_DEBUG(NE_DBG_HTTP, "Handled request; %d requests remain.\n",
-                 parms->num_requests);
+                 parms->num_requests - 1);
     } while (--parms->num_requests);
 
     return OK;
@@ -721,6 +780,8 @@ static int test_digest(struct digest_parms *parms)
 {
     ne_session *sess;
 
+    NE_DEBUG(NE_DBG_HTTP, ">>>> Request sequence begins.\n");
+
     if (parms->proxy) {
         sess = ne_session_create("http", "www.example.com", 80);
         ne_session_proxy(sess, "localhost", 7777);
@@ -746,24 +807,27 @@ static int digest(void)
 {
     struct digest_parms parms[] = {
         /* RFC 2617-style */
-        { "WallyWorld", "this-is-a-nonce", NULL, 1, 0, 0, 0, 0, 1, fail_not },
-        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", 1, 0, 0, 0, 0, 1, fail_not },
+        { "WallyWorld", "this-is-a-nonce", NULL, NULL, 1, 0, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", NULL, 1, 0, 0, 0, 0, 1, 0, fail_not },
         /* ... with A-I */
-        { "WallyWorld", "nonce-nonce-nonce", "opaque-string", 1, 1, 0, 0, 0, 1, fail_not },
+        { "WallyWorld", "nonce-nonce-nonce", "opaque-string", NULL, 1, 1, 0, 0, 0, 1, 0, fail_not },
         /* ... with md5-sess. */
-        { "WallyWorld", "nonce-nonce-nonce", "opaque-string", 1, 1, 1, 0, 0, 1, fail_not },
+        { "WallyWorld", "nonce-nonce-nonce", "opaque-string", NULL, 1, 1, 1, 0, 0, 1, 0, fail_not },
         /* many requests, with changing nonces; tests for next-nonce handling bug. */
-        { "WallyWorld", "this-is-a-nonce", "opaque-thingy", 1, 1, 0, 0, 1, 20, fail_not },
+        { "WallyWorld", "this-is-a-nonce", "opaque-thingy", NULL, 1, 1, 0, 0, 1, 20, 0, fail_not },
+        /* staleness. */
+        { "WallyWorld", "this-is-a-nonce", "opaque-thingy", NULL, 1, 1, 0, 0, 0, 3, 2, fail_not },
+
 
         /* RFC 2069-style */ 
-        { "WallyWorld", "lah-di-da-di-dah", NULL, 0, 0, 0, 0, 0, 1, fail_not },
-        { "WallyWorld", "fee-fi-fo-fum", "opaque-string", 0, 0, 0, 0, 0, 1, fail_not },
-        { "WallyWorld", "fee-fi-fo-fum", "opaque-string", 0, 1, 0, 0, 0, 1, fail_not },
+        { "WallyWorld", "lah-di-da-di-dah", NULL, NULL, 0, 0, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "fee-fi-fo-fum", "opaque-string", NULL, 0, 0, 0, 0, 0, 1, 0, fail_not },
+        { "WallyWorld", "fee-fi-fo-fum", "opaque-string", NULL, 0, 1, 0, 0, 0, 1, 0, fail_not },
 
         /* Proxy auth */
-        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", 1, 1, 0, 0, 0, 1, fail_not },
+        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", NULL, 1, 1, 0, 0, 0, 1, 0, fail_not },
         /* Proxy + A-I */
-        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", 1, 1, 0, 1, 0, 1, fail_not },
+        { "WallyWorld", "this-is-also-a-nonce", "opaque-string", NULL, 1, 1, 0, 1, 0, 1, 0, fail_not },
 
         { NULL }
     };
@@ -785,24 +849,25 @@ static int digest_failures(void)
         const char *message;
     } fails[] = {
         { fail_ai_bad_nc, "nonce count mismatch" },
+        { fail_ai_bad_nc_syntax, "could not parse nonce count" },
         { fail_ai_bad_digest, "digest mismatch" },
         { fail_ai_bad_cnonce, "client nonce mismatch" },
         { fail_ai_omit_nc, "missing parameters" },
         { fail_ai_omit_digest, "missing parameters" },
         { fail_ai_omit_cnonce, "missing parameters" },
-        { fail_bogus_alg, "Unknown algorithm" },
+        { fail_bogus_alg, "unknown algorithm" },
+        { fail_req0_stale, "initial Digest challenge was stale" },
         { fail_not, NULL }
     };
     size_t n;
 
+    memset(&parms, 0, sizeof parms);
+    
     parms.realm = "WallyWorld";
     parms.nonce = "random-invented-string";
     parms.opaque = NULL;
     parms.rfc2617 = 1;
     parms.send_ainfo = 1;
-    parms.md5_sess = 0;
-    parms.proxy = 0;
-    parms.send_nextnonce = 0;
     parms.num_requests = 1;
 
     for (n = 0; fails[n].message; n++) {
@@ -825,7 +890,8 @@ static int digest_failures(void)
 
         ne_session_destroy(sess);
         
-        if (fails[n].mode == fail_bogus_alg) {
+        if (fails[n].mode == fail_bogus_alg
+            || fails[n].mode == fail_req0_stale) {
             reap_server();
         } else {
             CALL(await_server());
@@ -835,8 +901,6 @@ static int digest_failures(void)
     return OK;
 }
 
-/* test that digest has precedence over Basic for multi-scheme
- * challenges */
 static int fail_cb(void *userdata, const char *realm, int tries, 
 		   char *un, char *pw)
 {
@@ -852,12 +916,41 @@ static int fail_cb(void *userdata, const char *realm, int tries,
 static int fail_challenge(void)
 {
     static const struct {
-        const char *resp, *challs;
+        const char *resp, *error, *challs;
     } ts[] = {
+        /* only possible Basic parse failure. */
+        { "Basic", "missing realm in Basic challenge" },
+
+        /* Digest parameter invalid/omitted failure cases: */
+        { "Digest algorithm=MD5, qop=auth, nonce=\"foo\"",
+          "missing parameter in Digest challenge" },
+        { "Digest algorithm=MD5, qop=auth, realm=\"foo\"",
+          "missing parameter in Digest challenge" },
+        { "Digest algorithm=ZEBEDEE-GOES-BOING, qop=auth, realm=\"foo\"",
+          "unknown algorithm in Digest challenge" },
+        { "Digest algorithm=MD5-sess, realm=\"foo\"",
+          "incompatible algorithm in Digest challenge" },
+        { "Digest algorithm=MD5, qop=auth, nonce=\"foo\", realm=\"foo\", "
+          "domain=\"http://[::1/\"", "could not parse domain" },
+
+        /* Multiple challenge failure cases: */
+        { "Basic, Digest",
+          "missing parameter in Digest challenge, missing realm in Basic challenge" },
+        
+        { "Digest realm=\"foo\", algorithm=MD5, qop=auth, nonce=\"foo\","
+          " Basic realm=\"foo\"",
+          "rejected Digest challenge, rejected Basic challenge" },
+
+        { "WhizzBangAuth realm=\"foo\", " 
+          "Basic realm='foo'",
+          "ignored WhizzBangAuth challenge, rejected Basic challenge" },
+        { "", "could not parse challenge" },
+
         /* neon 0.26.x regression in "attempt" handling. */
         { "Basic realm=\"foo\", " 
-          "Digest realm=\"bar\", algorithm=MD5, qop=auth, nonce=\"foo\"", 
-          "<bar, 0><foo, 1>"  /* Digest challenge first, Basic second. */
+          "Digest realm=\"bar\", algorithm=MD5, qop=auth, nonce=\"foo\"",
+          "rejected Digest challenge, rejected Basic challenge"
+          , "<bar, 0><foo, 1>"  /* Digest challenge first, Basic second. */
         }
     };
     unsigned n;
@@ -879,10 +972,18 @@ static int fail_challenge(void)
         ne_set_server_auth(sess, fail_cb, buf);
         
         ret = any_2xx_request(sess, "/fish");
-        ONN("request success", ret == NE_OK);
+        ONV(ret == NE_OK,
+            ("request success; expecting error '%s'",
+             ts[n].error));
+
+        ONV(strstr(ne_get_error(sess), ts[n].error) == NULL,
+            ("request fails with error '%s'; expecting '%s'",
+             ne_get_error(sess), ts[n].error));
         
-        ONCMP(ts[n].challs, buf->data, "challenge callback", 
-              "invocation order");
+        if (ts[n].challs) {
+            ONCMP(ts[n].challs, buf->data, "challenge callback", 
+                  "invocation order");
+        }
 
         ne_session_destroy(sess);
         ne_buffer_destroy(buf);
@@ -901,11 +1002,9 @@ static int multi_cb(void *userdata, const char *realm, int tries,
                     char *un, char *pw)
 {
     struct multi_context *ctx = userdata;
-    char buf[128];
-    
-    ne_snprintf(buf, sizeof buf, "[id=%d, realm=%s, tries=%d]", 
-                ctx->id, realm, tries);
-    ne_buffer_zappend(ctx->buf, buf);
+
+    ne_buffer_snprintf(ctx->buf, 128, "[id=%d, realm=%s, tries=%d]", 
+                       ctx->id, realm, tries);
 
     return -1;
 }
@@ -943,6 +1042,79 @@ static int multi_handler(void)
     return await_server();
 }
 
+static int domains(void)
+{
+    ne_session *sess;
+    struct digest_parms parms;
+
+    memset(&parms, 0, sizeof parms);
+    parms.realm = "WallyWorld";
+    parms.rfc2617 = 1;
+    parms.nonce = "agoog";
+    parms.domain = "http://localhost:7777/fish/ https://example.com /agaor /other";
+    parms.num_requests = 6;
+
+    CALL(make_session(&sess, serve_digest, &parms));
+
+    ne_set_server_auth(sess, auth_cb, NULL);
+
+    ne_session_proxy(sess, "localhost", 7777);
+
+    CALL(any_2xx_request(sess, "/fish/0"));
+    CALL(any_2xx_request(sess, "/outside"));
+    CALL(any_2xx_request(sess, "/others"));
+    CALL(any_2xx_request(sess, "/fish"));
+    CALL(any_2xx_request(sess, "/fish/2"));
+    CALL(any_2xx_request(sess, "*"));
+    
+    ne_session_destroy(sess);
+
+    return await_server();
+}
+
+/* This segfaulted with 0.28.0 through 0.28.2 inclusive. */
+static int CVE_2008_3746(void)
+{
+    ne_session *sess;
+    struct digest_parms parms;
+
+    memset(&parms, 0, sizeof parms);
+    parms.realm = "WallyWorld";
+    parms.rfc2617 = 1;
+    parms.nonce = "agoog";
+    parms.domain = "foo";
+    parms.num_requests = 1;
+
+    CALL(make_session(&sess, serve_digest, &parms));
+
+    ne_set_server_auth(sess, auth_cb, NULL);
+
+    ne_session_proxy(sess, "localhost", 7777);
+
+    any_2xx_request(sess, "/fish/0");
+    
+    ne_session_destroy(sess);
+
+    return await_server();
+}
+
+static int defaults(void)
+{
+    ne_session *sess;
+    
+    CALL(make_session(&sess, auth_serve, CHAL_WALLY));
+    ne_add_server_auth(sess, NE_AUTH_DEFAULT, auth_cb, NULL);
+    CALL(any_2xx_request(sess, "/norman"));
+    ne_session_destroy(sess);
+    CALL(await_server());
+
+    CALL(make_session(&sess, auth_serve, CHAL_WALLY));
+    ne_add_server_auth(sess, NE_AUTH_ALL, auth_cb, NULL);
+    CALL(any_2xx_request(sess, "/norman"));
+    ne_session_destroy(sess);
+    return await_server();
+}
+
 /* test logout */
 
 /* proxy auth, proxy AND origin */
@@ -958,5 +1130,8 @@ ne_test tests[] = {
     T(digest_failures),
     T(fail_challenge),
     T(multi_handler),
+    T(domains),
+    T(defaults),
+    T(CVE_2008_3746),
     T(NULL)
 };
